package ca.recrown.islandsurvivalcraft.world.generation;

import org.bukkit.Chunk;
import org.bukkit.World;
import org.bukkit.block.Biome;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.world.ChunkLoadEvent;

import ca.recrown.islandsurvivalcraft.Utilities;
import ca.recrown.islandsurvivalcraft.Types.Point2;
import ca.recrown.islandsurvivalcraft.caching.Cache;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateTargetValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.DepthFirstSearch;
import ca.recrown.islandsurvivalcraft.world.BiomeSelector;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMapper;

//Note: technically, the validators have to be run on land, and so, some condition checks may not be nessecary.
public class BiomePerIslandGenerator implements IslandBiomeGenerator {
    private volatile boolean initialized;
    private final TemperatureMapGenerator temperatureMapGenerator;
    private final Cache<Point2, Biome> chunkBiomesCache;
    private final Cache<Point2, Boolean> chunkGenStatusCache;
    private volatile IslandWorldMapper worldIslandMap;
    private volatile BiomeSelector biomeSelector;
    private volatile World world;
    private volatile Point2 currChunkCoords;
    private volatile Biome[][] localChunkCache;
    
    float temperature;

    public BiomePerIslandGenerator() {
        this.temperatureMapGenerator = new TemperatureMapGenerator();
        chunkBiomesCache = new Cache<>(65536);
        chunkGenStatusCache = new Cache<>(16384);
        localChunkCache = new Biome[16][16];
    }
    
    @Override
    public void initialize(final World world, final IslandWorldMapper mapGenerator, final BiomeSelector biomeSelector) {
        if (initialized) throw new IllegalStateException("Biome generator already initialized.");
        initialized = true;
        this.world = world;
        this.worldIslandMap = mapGenerator;
        this.biomeSelector = biomeSelector;
        this.temperatureMapGenerator.setSeed(world.getSeed());
    }

    @EventHandler(priority = EventPriority.HIGH)
    public void onChunkLoaded(ChunkLoadEvent event) {
        if (event.isNewChunk()) {
            Chunk chunk = event.getChunk();
            if (chunk.getWorld().getUID().hashCode() == world.getUID().hashCode() && chunk.getWorld().getUID().equals(world.getUID())) {
                chunkGenStatusCache.setValue(new Point2(chunk.getX(), chunk.getZ()), true);
            }
        }
    }

    @Override
    public IslandBiomeGenerator getInstance() {
        return new BiomePerIslandGenerator();
    }

    @Override
    public Biome GenerateBiome(int chunkX, int chunkZ, int localX, int localZ) {
        if (currChunkCoords == null || chunkX != currChunkCoords.x || chunkZ != currChunkCoords.y) {
            currChunkCoords = new Point2(chunkX, chunkZ);
            localChunkCache = new Biome[16][16];
        }
        int worldX = Utilities.addMagnitude(16 * chunkX, localX);
        int worldZ = Utilities.addMagnitude(16 * chunkZ, localZ);
        
        Point2 worldCoords = new Point2(worldX, worldZ);

        Biome cachedBiome = getBiome(worldCoords);
        if (cachedBiome != null) return cachedBiome;
        
        temperature = temperatureMapGenerator.getTemperature(worldX, worldZ);
        if (!worldIslandMap.isIsland(worldX, worldZ)) {
            return biomeSelector.getOceanBiome(temperature);
        }

        IslandBiomeInfo biomeInfo = new IslandBiomeInfo();
        DepthFirstSearch prevBiomeFinder = new DepthFirstSearch(biomeInfo);
        prevBiomeFinder.setStartPosition(worldX, worldZ);
        if (!prevBiomeFinder.findTarget(biomeInfo)) {
            if (biomeInfo.main == null) biomeInfo.main = biomeSelector.getLandBiome(temperature);
            if (biomeInfo.shore == null) biomeInfo.shore = biomeSelector.getShoreBiome(biomeInfo.main, temperature);
            if (biomeInfo.shallow == null) biomeInfo.shallow = biomeSelector.getOceanBiome(temperature);            
        }
        FreshCachePropagationInfo propInfo = new FreshCachePropagationInfo(biomeInfo.main, biomeInfo.shore, biomeInfo.shallow);
        DepthFirstSearch prop = new DepthFirstSearch(propInfo);
        prop.setStartPosition(worldX, worldZ);
        prop.findTarget(propInfo);
        return getBiome(worldCoords);
    }

    private Biome getBiome(Point2 worldCoords) {
        int localX = Math.abs(worldCoords.x % 16);
        int localZ = Math.abs(worldCoords.y % 16);
        Point2 chunkCoords = new Point2(worldCoords.x / 16, worldCoords.y / 16);

        //chunk local
        if (chunkCoords.fastEquals(this.currChunkCoords) && localChunkCache[localX][localZ] != null) {
            return localChunkCache[localX][localZ];
        }

        //cache local
        Biome biome = chunkBiomesCache.getValue(worldCoords);
        if (biome != null) return biome;

        //files local
        Boolean chunkGenStat = chunkGenStatusCache.getValue(chunkCoords);
        if (chunkGenStat == null) {
            chunkGenStat = world.isChunkGenerated(chunkCoords.x, chunkCoords.y);
            chunkGenStatusCache.setValue(chunkCoords, chunkGenStat);
        }
        
        if (chunkGenStat) {
            biome = world.getBiome(worldCoords.x, 0, worldCoords.y);
            chunkBiomesCache.setValue(worldCoords, biome);
            return biome;
        }
        return null;
    }

    private void setCacheBiome(Point2 worldCoords, Biome biome) {
        int localX = Math.abs(worldCoords.x % 16);
        int localZ = Math.abs(worldCoords.y % 16);
        Point2 chunkCoords = new Point2(worldCoords.x / 16, worldCoords.y / 16);

        //If local, set it locally.
        if (chunkCoords.fastEquals(this.currChunkCoords)) localChunkCache[localX][localZ] = biome;

        //Set it cache wide.
        chunkBiomesCache.setValue(worldCoords, biome);
    }

    private class FreshCachePropagationInfo implements CoordinateTargetValidatable, CoordinateValidatable {
        public final Biome main, shore, shallow;

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            Point2 worldCoords = new Point2(x, y);
            if (worldIslandMap.isLand(x, y)) {
                if (worldIslandMap.isShore(x, y)) {
                    setCacheBiome(worldCoords, shore);
                } else {
                    setCacheBiome(worldCoords, main);
                }
            } else {
                setCacheBiome(worldCoords, shallow);
            }
            return false;
        }
        
        @Override
        public boolean validate(int x, int y) {
            Point2 chunkCoords = new Point2(x / 16, y / 16);
            return chunkCoords.fastEquals(currChunkCoords) && worldIslandMap.isIsland(x, y);
        }

        public FreshCachePropagationInfo(Biome main, Biome shore, Biome shallow) {
            this.main = main;
            this.shore = shore;
            this.shallow = shallow;
        }
    }

    private class IslandBiomeInfo implements CoordinateTargetValidatable, CoordinateValidatable {
        private boolean isLand, isShore, isShallow;
        public Biome main, shore, shallow;
        
        @Override
        public boolean validate(int x, int y) {
            isShore = false;
            isShallow = false;
            isLand = false;
            return worldIslandMap.isIsland(x, y);
        }

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            Point2 worldCoords = new Point2(x, y);

            if ((isLand = worldIslandMap.isLand(x, y))) {
                isShore = worldIslandMap.isShore(x, y);
            } else {
                isShallow = worldIslandMap.isShallowPortion(x, y);
            }

            if (main == null && isLand && !isShore) {
                main = getBiome(worldCoords);
            } else
            if (shore == null && isShore) {
                shore = getBiome(worldCoords);
            } else
            if (shallow == null && isShallow) {
                shallow = getBiome(worldCoords);
            }
            return allBiomesAcquired();
        }

        public boolean allBiomesAcquired() {
            return shore != null && main != null && shallow != null;
        }
    }

    
}